import os
import pickle

import requests.cookies

import venmo


def save(requests_cookiejar):
    try:
        os.makedirs(os.path.dirname(venmo.settings.COOKIES_FILE))
    except OSError:
        pass  # It's okay if directory already exists
    with open(venmo.settings.COOKIES_FILE, 'wb') as f:
        pickle.dump(requests_cookiejar, f, -1)


def load():
    try:
        with open(venmo.settings.COOKIES_FILE, 'rb') as f:
            return pickle.load(f)
    except IOError:
        # No cookies
        return requests.cookies.RequestsCookieJar()



import logging

from ._version import __version__  # noqa

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(name)s: %(levelname)s %(message)s')

logging.getLogger('requests').setLevel(logging.WARNING)


from . import (  # noqa: F401
    auth,
    cli,
    cookies,
    payment,
    settings,
    singletons,
    types,
    user
)



import atexit

from requests import Session

import venmo

_session = None


def _save_cookies():
    venmo.cookies.save(session().cookies)


def session():
    global _session
    if not _session:
        _session = Session()
        _session.cookies = venmo.cookies.load()
        atexit.register(_save_cookies)
    return _session



'''
User module.
'''
import json

import requests

import venmo


def id_from_username(username):
    for u in search(username):
        if u['username'].lower() == username.lower():
            return u['id']
    return None


def print_search(query):
    print(json.dumps(search(query), indent=4))


def search(query):
    response = requests.get(
        venmo.settings.USERS_URL,
        params={
            'limit': 5,
            'query': query,
        }
    )
    users = response.json()['data']
    results = []
    for u in users:
        results.append({
            'id': u['id'],
            'username': u['username'],
            'display_name': u['display_name'],
            'profile_picture_url': u['profile_picture_url'],
        })
    return results



'''
Authentication
'''

import getpass
import logging
import os
import os.path
import re
import shutil
import xml.etree.ElementTree as ET

import venmo

# Python 2.x fixes
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

try:
    # pylint: disable=redefined-builtin
    input = raw_input
except NameError:
    pass


logger = logging.getLogger('venmo.auth')


def configure():
    '''Get venmo ready to make payments.

    First, set username and password. If those change, then get access token.
    If that fails, unset username and password.

    Return whether or not we save an access token.
    '''
    # Update credentials
    credentials = update_credentials()
    if not credentials:
        return False
    else:
        email, password = credentials

    # Log in to Auto
    success = submit_credentials(email, password)
    if not success:
        return False
    else:
        redirect_url, csrftoken2 = success

    # 2FA is expected because issue #23
    if 'two-factor' not in redirect_url:
        logger.error('invalid credentials')
        return False

    # Write email password
    config = read_config()
    config.set(configparser.DEFAULTSECT, 'email', email)
    config.set(configparser.DEFAULTSECT, 'password', password)
    write_config(config)

    # Do 2FA
    access_token = two_factor(redirect_url, csrftoken2, email, password)
    if not access_token:
        return False

    # Write access token
    config = read_config()
    config.set(configparser.DEFAULTSECT, 'access_token', access_token)
    write_config(config)
    return True


def two_factor(redirect_url, csrftoken2, email, password):
    '''Do the two factor auth dance.

    Return access_token or False.
    '''
    logger.info('Sending SMS verification ...')

    # Get two factor page
    response = venmo.singletons.session().get(redirect_url)

    # Send SMS
    secret = extract_otp_secret(response.text)
    headers = {'Venmo-Otp-Secret': secret}
    data = {
        'via': 'sms',
        'csrftoken2': csrftoken2,
    }
    response = venmo.singletons.session().post(
        venmo.settings.TWO_FACTOR_URL,
        json=data,
        headers=headers,
    )
    assert response.status_code == 200, 'Post to 2FA failed'
    assert response.json()['data']['status'] == 'sent', 'SMS did not send'

    # Prompt verification code
    verification_code = input('Verification code: ')
    if not verification_code:
        logger.error('verification code required')
        return False

    # Submit verification code
    data = {
        'csrftoken2': csrftoken2,
        'return_json': 'true',
        'password': password,
        'phoneEmailUsername': email,
        'token': verification_code,
    }
    response = venmo.singletons.session().post(
        venmo.settings.TWO_FACTOR_AUTHORIZATION_URL,
        json=data,
        allow_redirects=False,
    )
    if response.status_code != 200:
        logger.error('verification code failed')
        return False

    # Retrieve access token
    access_token = response.json()['access_token']
    return access_token


def extract_otp_secret(text):
    pattern = re.compile(r'"secret":"(\w*)"')
    for line in text.splitlines():
        match = pattern.search(line)
        return match.group(1)
    raise Exception('msg="Could not extract data-otp-secret"')


def retrieve_access_token(code):
    data = {
        'client_id': venmo.settings.CLIENT_ID,
        'client_secret': venmo.settings.CLIENT_SECRET,
        'code': code,
    }
    response = venmo.singletons.session().post(venmo.settings.ACCESS_TOKEN_URL,
                                               data)
    response_dict = response.json()
    access_token = response_dict['access_token']
    return access_token


def _authorization_url():
    scopes = [
        'make_payments',
        'access_feed',
        'access_profile',
        'access_email',
        'access_phone',
        'access_balance',
        'access_friends',
    ]
    params = {
        'client_id': venmo.settings.CLIENT_ID,
        'scope': ' '.join(scopes),
        'response_type': 'code',
    }
    return '{authorization_url}?{params}'.format(
        authorization_url=venmo.settings.AUTHORIZATION_URL,
        params=urlencode(params)
    )


def _filter_tag(input_xml, tag):
    '''Filter out the script so we can parse the xml.'''
    output_lines = []
    inside = False
    for line in input_xml.splitlines():
        if '<{tag}>'.format(tag=tag) in line:
            inside = True
        if not inside:
            output_lines.append(line)
        if '</{tag}>'.format(tag=tag) in line:
            inside = False
    return '\n'.join(output_lines)


def update_credentials():
    '''Save username and password to config file.

    Entering nothing keeps the current credentials. Returns whether or not
    the credentials changed.
    '''
    # Read old credentials
    config = read_config()
    try:
        old_email = config.get(configparser.DEFAULTSECT, 'email')
    except configparser.NoOptionError:
        old_email = ''
    try:
        old_password = config.get(configparser.DEFAULTSECT, 'password')
    except configparser.NoOptionError:
        old_password = ''

    # Prompt new credentials
    email = input('Venmo email [{}]: '
                  .format(old_email if old_email else None))
    password = getpass.getpass(prompt='Venmo password [{}]: '
                               .format('*' * 10 if old_password else None))
    email = email or old_email
    password = password or old_password

    incomplete = not email or not password
    if incomplete:
        logger.warn('credentials incomplete')
        return False

    return email, password


def submit_credentials(email, password):
    # Get and parse authorization webpage xml and form
    response = venmo.singletons.session().get(_authorization_url())
    authorization_page_xml = response.text
    filtered_xml = _filter_tag(authorization_page_xml, 'script')
    filtered_xml = _filter_tag(filtered_xml, 'head')
    root = ET.fromstring(filtered_xml)
    form = root.find('.//form')
    for child in form:
        if child.attrib.get('name') == 'csrftoken2':
            csrftoken2 = child.attrib['value']
        if child.attrib.get('name') == 'auth_request':
            auth_request = child.attrib['value']
        if child.attrib.get('name') == 'web_redirect_url':
            web_redirect_url = child.attrib['value']

    # Submit form
    data = {
        'username': email,
        'password': password,
        'web_redirect_url': web_redirect_url,
        'csrftoken2': csrftoken2,
        'auth_request': auth_request,
        'grant': 1,
    }
    response = venmo.singletons.session().post(
        venmo.settings.AUTHORIZATION_URL,
        json=data,
        allow_redirects=False,
    )

    if response.status_code != 302:
        logger.error('expecting a redirect')
        return False

    redirect_url = response.headers['location']
    return redirect_url, csrftoken2


def get_username():
    config = read_config()
    try:
        return config.get(configparser.DEFAULTSECT, 'email')
    except configparser.NoOptionError:
        return None


def get_password():
    config = read_config()
    return config.get(configparser.DEFAULTSECT, 'password')


def get_access_token():
    config = read_config()
    try:
        return config.get(configparser.DEFAULTSECT, 'access_token')
    except configparser.NoOptionError:
        return None


def read_config():
    config = configparser.RawConfigParser()
    config.read(venmo.settings.CREDENTIALS_FILE)
    return config


def write_config(config):
    try:
        os.makedirs(os.path.dirname(venmo.settings.CREDENTIALS_FILE))
    except OSError:
        pass  # It's okay if directory already exists
    with open(venmo.settings.CREDENTIALS_FILE, 'w') as configfile:
        config.write(configfile)


def reset():
    '''rm -rf ~/.venmo'''
    shutil.rmtree(venmo.settings.DOT_VENMO)



#!/usr/bin/env python

'''
Venmo CLI.

Pay or charge people via the Venmo API:

  venmo pay @zackhsi 23.19 'Thanks for the beer <3'
  venmo charge 19495551234 23.19 'That beer wasn't free!'
'''

import argparse
import os
import sys
from datetime import datetime

import venmo


def status():
    '''Print out system status

    $ venmo status
    Version 0.3.2
    Credentials (updated 2016-01-26 19:48):
        User: youremailaddress
        Token: youraccesstoken
    '''
    print('\n'.join([_version(), _credentials()]))


def _version():
    return 'Version {}'.format(venmo.__version__)


def _credentials():
    try:
        updated_at = os.path.getmtime(venmo.settings.CREDENTIALS_FILE)
        updated_at = datetime.fromtimestamp(updated_at)
        updated_at = updated_at.strftime('%Y-%m-%d %H:%M')
        return '''Credentials (updated {updated_at}):
    User: {user}
    Token: {token}'''.format(updated_at=updated_at,
                             user=venmo.auth.get_username(),
                             token=venmo.auth.get_access_token())
    except OSError:
        return 'No credentials'


def parse_args():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers()

    for action in ['pay', 'charge']:
        subparser = subparsers.add_parser(action,
                                          help='{} someone'.format(action))
        subparser.add_argument(
            'user',
            help='who to {}, either phone or username'.format(action),
        )
        subparser.add_argument('amount', type=venmo.types.positive_float,
                               help='how much to pay or charge')
        subparser.add_argument('note', help='what the request is for')
        subparser.set_defaults(func=getattr(venmo.payment, action))

    parser_configure = subparsers.add_parser('configure',
                                             help='set up credentials')
    parser_configure.set_defaults(func=venmo.auth.configure)

    parser_search = subparsers.add_parser('search', help='search users')
    parser_search.add_argument('query', help='search query')
    parser_search.set_defaults(func=venmo.user.print_search)

    parser_status = subparsers.add_parser('status', help='get status')
    parser_status.set_defaults(func=status)

    parser_reset = subparsers.add_parser('reset', help='reset saved data')
    parser_reset.set_defaults(func=venmo.auth.reset)

    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + venmo.__version__)

    if len(sys.argv) == 1:
        sys.argv.append('-h')
    args = parser.parse_args()
    func = args.func
    del args.func
    func(**vars(args))


def main():
    try:
        parse_args()
    except KeyboardInterrupt:
        print('')


if __name__ == '__main__':
    main()



from setuptools import setup

# 'Import' __version__
exec(open('venmo/_version.py').read())


setup(
    name='venmo',
    version=__version__,  # noqa
    description='Venmo CLI',
    url='http://github.com/zackhsi/venmo',
    author='Zack Hsi',
    author_email='zackhsi@gmail.com',
    license='MIT',
    packages=['venmo'],
    install_requires=[
        'requests>=2.9.1',
    ],
    entry_points={
        'console_scripts': [
            'venmo = venmo.cli:main',
        ],
    },
)



import os


# OAuth
CLIENT_ID = '2667'
CLIENT_SECRET = 'srDrmU3yf452HuFF63HqHEt25pa5DexZ'

# Paths
DOT_VENMO = os.path.join(os.path.expanduser('~'), '.venmo')
CREDENTIALS_FILE = os.path.join(DOT_VENMO, 'credentials')
COOKIES_FILE = os.path.join(DOT_VENMO, 'cookies')

# URLs
ACCESS_TOKEN_URL = 'https://api.venmo.com/v1/oauth/access_token'
AUTHORIZATION_URL = 'https://api.venmo.com/v1/oauth/authorize'
PAYMENTS_URL = 'https://api.venmo.com/v1/payments'
TWO_FACTOR_URL = 'https://venmo.com/api/v5/two_factor/token'
TWO_FACTOR_AUTHORIZATION_URL = 'https://venmo.com/login'
USERS_URL = 'https://api.venmo.com/v1/users'



'''
Payment
'''

import logging
import sys
import requests

import venmo

logger = logging.getLogger('venmo.payment')


def pay(user, amount, note):
    _pay_or_charge(user, amount, note)


def charge(user, amount, note):
    amount = -amount
    _pay_or_charge(user, amount, note)


def _pay_or_charge(user, amount, note):
    access_token = venmo.auth.get_access_token()
    if not access_token:
        logger.warn('No access token. Configuring ...')
        if not venmo.auth.configure():
            return
        access_token = venmo.auth.get_access_token()

    data = {
        'note': note,
        'amount': amount,
        'access_token': access_token,
        'audience': 'private',
    }
    if user.startswith('@'):
        username = user[1:]
        user_id = venmo.user.id_from_username(username)
        if not user_id:
            logger.error('Could not find user @{}'.format(username))
            return
        data['user_id'] = user_id
    else:
        data['phone'] = user

    response = requests.post(
        venmo.settings.PAYMENTS_URL,
        json=data
    )
    data = response.json()

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        error_message = 'received {} from Venmo'.format(e.response.status_code)
        if 'error' in data:
            message = data['error']['message']
            error_message += ': "{}"'.format(message)
        print(error_message)
        sys.exit(1)

    payment = data['data']['payment']
    target = payment['target']
    payment_action = payment['action']
    if payment_action == 'charge':
        payment_action = 'charged'
    if payment_action == 'pay':
        payment_action = 'paid'
    amount = payment['amount']
    if target['type'] == 'user':
        user = '{first_name} {last_name}'.format(
            first_name=target['user']['first_name'],
            last_name=target['user']['last_name'],
        )
    else:
        user = target[target['type']],
    note = payment['note']
    print('Successfully {payment_action} {user} ${amount:.2f} for "{note}"'
          .format(
              payment_action=payment_action,
              user=user,
              amount=amount,
              note=note,
          ))
